/*	$NetBSD: biosreboot.S,v 1.4 2005/12/11 12:17:48 christos Exp $	*/

/*
 * Copyright (c) 1997
 * 	Perry E. Metzger.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgements:
 *	This product includes software developed for the NetBSD Project
 *	by Perry E. Metzger.
 * 4. The names of the authors may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */


#include <machine/asm.h>

	.text

	.set BDA_BOOT,0x472		/* Boot howto flag location */

/*
 *	Reboot the system via the BIOS.
 */
ENTRY(reboot)
	pushl	%ebp
	movl	%esp,%ebp
	pushl	%ebx
	push	%esi
	push	%edi

	call	_C_LABEL(prot_to_real)
	.code16

	/* 
	 * We don't want to use INT 19 for reboot because it isn't
	 * guaranteed to work, and besides it is what the BIOS itself
	 * calls _after_ a successful completion of the POST routines.
	 *
	 * From the IBM PC BIOS documentation:
	 *
	 * "INT 19 - SYSTEM - BOOTSTRAP LOADER
	 * 
	 * "This interrupt reboots the system without clearing memory
	 * or restoring interrupt vectors.  Because interrupt vectors
	 * are preserved, this interrupt usually causes a system hang
	 * if any TSRs have hooked vectors from 00h through 1Ch,
	 * particularly INT 08.
	 * 
	 * "To accomplish a warm boot equivalent to Ctrl-Alt-Del,
	 * store 1234h in 0040h:0072h and jump to FFFFh:0000h.  For a
	 * cold boot equivalent to a reset, store 0000h at 0040h:0072h
	 * before jumping."
	 * 
	 * Note: Unfortunately both warm and cold boot on VirtualBox
	 * (since at least 4.0.4 through to at least 4.1.8) hang the
	 * (virtual) system.  What does seem to work on VirtualBox,
	 * from the kernel, is the PC/AT Control Port A Fast Reset.
	 * (PCI-bus system & CPU reset does not work.)
	 *
	 * From the Linux kernel:
	 *
	 * "Write zero to CMOS register number 0x0f, which the BIOS
	 * POST routine will recognize as telling it to do a proper
	 * reboot.  (Well that's what this book in front of me says --
	 * it may only apply to the Phoenix BIOS though, it's not
	 * clear).  At the same time, disable NMIs by setting the top
	 * bit in the CMOS address register, as we're about to do
	 * peculiar things to the CPU."
	 *
	 *	outb_p (0x8f, 0x70);
	 *	outb_p (0x00, 0x71);
	 * 
	 * For now we will try what the BIOS folks call a "cold boot"
	 * (it's actually a form of warm boot, since it does not
	 * trigger a hardware reset) since this will ensure the most
	 * complete reset of system state is done without actually
	 * causing a hardware reset.
	 */
	movw	$0x0000, BDA_BOOT	/* set cold boot (clear warm boot flag) */
	/*
	 * XXX apparently this would only be correct for an original 8086:
	 *
	ljmp	$0xffff, $0x0000
	 */
	/*
	 * "however in various literature it is mentioned that
	 * $0xf000,$0xfff0 is bound to work better..."
	 *
	 *	http://unix.derkeiler.com/Mailing-Lists/FreeBSD/hackers/2003-11/0205.html
	 * 
	 * "Apparently CPUs haven't started with the CS:IP combination
	 * FFFF:0000 since the 8086.  The 80286 changed this to
	 * F000:FFF0.
	 * 
	 * "32-bit CPUs start up in what is colloquially known as
	 * unreal mode.  Even though the initial value of the CS
	 * register after reset is F000, the segment descriptor
	 * associated with that register initially holds FFFF0000 as
	 * its base address.  So the physical address that initially
	 * corresponds to the the 16:16 CS:IP address F000:FFF0 is in
	 * fact, and has been since the days of the 80386, FFFFFFF0.
	 * And that's where the machine firmware is principally mapped
	 * into physical address space on 32-bit and 64-bit x86
	 * machines."
	 * 
	 *	http://superuser.com/questions/336021/is-bios-read-from-the-bios-chip-or-copied-into-ram-on-startup
	 * 
	 * Since we are in 16-bit "real" mode, we can use the 16:16
	 * address here.
	 * 
	 * There's also a story that one should never try to reset the
	 * CPU without enabling CPU A20 to the address BUS.
	 * 
	 *	http://www.rcollins.org/Productivity/A20Reset.html
	 * 
	 * See here for doing this:
	 * 
	 *	http://www.brokenthorn.com/Resources/OSDev9.html
	 * 
	 * However others say "the CPU restart vector on 286 and above
	 * processors is not F000:FFF0.  These CPUs actually execute
	 * out of the very top of their physical address space, which
	 * in some cases is occupied by a [special] boot loader...
	 * Even though the CPU is executing out of memory above the
	 * 1MB address mark, it is still executing in real mode, not
	 * protected mode (really, real mode). Once the CS register is
	 * reloaded with any value, the CPU disables the upper address
	 * lines and, typically, continues to execute at the
	 * 8086-compatible reboot address, F000:FFF0".
	 * 
	 * NOTE:  Apparently on some machines there will be an INT 19h
	 * instruction here, while others will jump to the main BIOS
	 * entry point (e.g. f000:e05b).
	 * 
	 * Jump to the BIOS "Power On Reset Vector":
	 */
	ljmp	$0xf000, $0xfff0

	/* NOTE: We will never get here.... */

	calll	_C_LABEL(real_to_prot)
	.code32

	movl	%ebx, %eax

	pop	%edi
	pop	%esi
	popl	%ebx
	popl	%ebp
	ret
